library simprokcore;

import 'package:simprokmachine/simprokmachine.dart';

void runRootCore<State>({
  required State initial,
  required WidgetLayer<State> main,
  required Set<MachineLayer<State>> secondary,
}) {
  final Machine<_StateAction<State>, _StateAction<State>> reducer =
      _ReducerMachine<Mapper<State, State>, State>.create(
    initial: initial,
    reducer: (state, event) => event(state),
  ).inward((_StateAction<State> input) {
    if (input.mapper != null) {
      // will update
      return Inward.Set(input.mapper!);
    } else {
      // did update
      return Inward.Nan();
    }
  }).outward(
    (State output) => Outward.Out(_StateAction<State>.didUpdate(output)),
  );

  final WidgetMachine<_StateAction<State>, _StateAction<State>> mainMachine =
      main._child();

  final Iterable<Machine<_StateAction<State>, _StateAction<State>>>
      secondaryMachines = secondary.map((MachineLayer<State> e) => e._child());

  final Set<Machine<_StateAction<State>, _StateAction<State>>> machines =
      List<Machine<_StateAction<State>, _StateAction<State>>>.from(
    secondaryMachines,
  ).toSet();
  machines.add(reducer);

  final WidgetMachine<_StateAction<State>, _StateAction<State>> root =
      mergeWidgetMachine(main: mainMachine, secondary: machines);

  runRootMachine<_StateAction<State>, _StateAction<State>>(
    root: root,
    initial: _StateAction<State>.willUpdate((State state) => state),
    map: (_StateAction<State> output) => Inward.Set(output),
  );
}

abstract class MachineLayer<GlobalState> {
  Machine<_StateAction<GlobalState>, _StateAction<GlobalState>> _child();

  Machine<_StateAction<GlobalState>, _StateAction<GlobalState>>
      _get<State, Event>(
    Machine<State, Event> machine,
    Mapper<GlobalState, State> mapper,
    BiMapper<GlobalState, Event, GlobalState> reducer,
  ) {
    return machine.inward(_layerInward(mapper)).outward(_layerOutward(reducer));
  }
}

abstract class MachineLayerType<GlobalState, State, Event>
    extends MachineLayer<GlobalState> {
  Machine<State, Event> machine();

  State map(GlobalState state);

  GlobalState reduce(GlobalState state, Event event);

  @override
  Machine<_StateAction<GlobalState>, _StateAction<GlobalState>> _child() {
    return _get(
      machine(),
      map,
      reduce,
    );
  }
}

class MachineLayerObject<GlobalState, State, Event>
    extends MachineLayer<GlobalState> {
  final Machine<State, Event> _machine;
  final Mapper<GlobalState, State> _mapper;
  final BiMapper<GlobalState, Event, GlobalState> _reducer;

  MachineLayerObject({
    required Machine<State, Event> machine,
    required Mapper<GlobalState, State> mapper,
    required BiMapper<GlobalState, Event, GlobalState> reducer,
  })  : _machine = machine,
        _mapper = mapper,
        _reducer = reducer;

  @override
  Machine<_StateAction<GlobalState>, _StateAction<GlobalState>> _child() {
    return _get(
      _machine,
      _mapper,
      _reducer,
    );
  }
}

abstract class WidgetLayer<GlobalState> {
  WidgetMachine<_StateAction<GlobalState>, _StateAction<GlobalState>>
      _get<State, Event>(
    WidgetMachine<State, Event> machine,
    Mapper<GlobalState, State> mapper,
    BiMapper<GlobalState, Event, GlobalState> reducer,
  ) {
    return machine.inward(_layerInward(mapper)).outward(_layerOutward(reducer));
  }

  WidgetMachine<_StateAction<GlobalState>, _StateAction<GlobalState>> _child();
}

abstract class WidgetLayerType<GlobalState, State, Event>
    extends WidgetLayer<GlobalState> {
  WidgetMachine<State, Event> machine();

  State map(GlobalState state);

  GlobalState reduce(GlobalState state, Event event);

  @override
  WidgetMachine<_StateAction<GlobalState>, _StateAction<GlobalState>> _child() {
    return _get(
      machine(),
      map,
      reduce,
    );
  }
}

class WidgetLayerObject<GlobalState, State, Event>
    extends WidgetLayer<GlobalState> {
  final WidgetMachine<State, Event> _machine;
  final Mapper<GlobalState, State> _mapper;
  final BiMapper<GlobalState, Event, GlobalState> _reducer;

  WidgetLayerObject({
    required WidgetMachine<State, Event> machine,
    required Mapper<GlobalState, State> mapper,
    required BiMapper<GlobalState, Event, GlobalState> reducer,
  })  : _machine = machine,
        _mapper = mapper,
        _reducer = reducer;

  @override
  WidgetMachine<_StateAction<GlobalState>, _StateAction<GlobalState>> _child() {
    return _get(
      _machine,
      _mapper,
      _reducer,
    );
  }
}

class _StateAction<T> {
  final T? state;
  final Mapper<T, T>? mapper;

  _StateAction.didUpdate(this.state) : mapper = null;

  _StateAction.willUpdate(this.mapper) : state = null;
}

class _ReducerMachine<Event, State> extends ChildMachine<Event, State> {
  State _initial;
  final BiMapper<State, Event, State> _reducer;

  _ReducerMachine.create({
    required State initial,
    required BiMapper<State, Event, State> reducer,
  })  : _initial = initial,
        _reducer = ((state, event) => reducer(state, event));

  @override
  void render(Event input, Handler<State> callback) {
    _initial = _reducer(_initial, input);
    callback(_initial);
  }
}

Mapper<_StateAction<GlobalState>, Inward<State>>
    _layerInward<GlobalState, State>(Mapper<GlobalState, State> mapper) {
  return (_StateAction<GlobalState> input) {
    if (input.state != null) {
      return Inward.Set(mapper(input.state!));
    } else {
      return Inward.Nan();
    }
  };
}

Mapper<Event, Outward<State, _StateAction<GlobalState>>>
    _layerOutward<GlobalState, State, Event>(
  BiMapper<GlobalState, Event, GlobalState> reducer,
) {
  return (Event output) {
    return Outward.Out(_StateAction.willUpdate(
      (GlobalState currentState) => reducer(currentState, output),
    ));
  };
}
